/*
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

var Utilities = function () { };

// The following variables are contained within their own local function scope...
(function () {

	var initComplete = false;
	var templateObj;

	// Wait for DOMContentLoaded event before trying to access the application template
	var init = function () {

		Utilities.getListOfUsbDeviceKeys = function (filterKey) {

			// Get a reference to 'ti-service-usb'
			let serviceUsb = gc.services['ti-service-usb'];
			if (!serviceUsb) {
				console.error("Could not find a reference to 'ti-service-usb'!");
			}

			// Get a list of matching device keys (formatted as "<vid>_<pid>")
			let usbDevices = Array.from(serviceUsb.deviceInfoMap.keys());
			if (filterKey) {
				usbDevices = usbDevices.filter((x) => (x.indexOf(filterKey) >= 0));
			}

			return usbDevices;
		};

		Utilities.getListOfBulkDeviceKeys = function () {
			const keyPrefix = "1cbe_00ab";          // SCB
			//const keyPrefix = "1cbe_0003";        // TIVA
			//const keyPrefix = "1cbe_af81";        // USB-PHI
			let bulkDevices = this.getListOfUsbDeviceKeys(keyPrefix);
			return bulkDevices;
		};

		Utilities.getListOfDfuDeviceKeys = function () {
			const DFU_deviceKey = "1cbe_00ff";		// MSP432E DFU Mode
			let dfuDevices = this.getListOfUsbDeviceKeys(DFU_deviceKey);
			return dfuDevices;
		};


		//
		// Safely access deeply nested values from JS objects
		//

		// Reference: https://dev.to/tbetous/how-to-safely-access-deep-nested-values-in-javascript-5b2a
		Utilities.getPropValue = (object, path = '') => path.split('.').reduce((o, x) => o == undefined ? o : o[x], object);

		// Reference: https://stackoverflow.com/questions/14782232/how-to-avoid-cannot-read-property-of-undefined-errors/42349521
		Utilities.getSafe = (fn, defaultVal) => {
			try {
				return fn();
			} catch (e) {
				return defaultVal;
			}
		};

		// Example relative path string: '/target/release_notes.txt';
		Utilities.fetchTextFile = async function (relPath) {

			//const appURL = location.href;
			//let fileURL = appURL.substring(0, appURL.lastIndexOf("/")) + relPath;
			let fileURL = Polymer.ResolveUrl.resolveUrl('.') + relPath;

			let response = await fetch(fileURL);
			let text = await response.text();

			return text;
		};

		Utilities.installDfuDriver = function () {

			// A helper function to install through addByCategory, categories is an array,
			// e.g ['categoryName1', 'categoryName2', ...]
			function addCategories(targetSupport, categories) {

				let progresshdl = function (e) {
					gc.console.debug('addCategories', e.subActivity, e.isComplete);
				};
				targetSupport.addListener('progress', progresshdl);
				//gc.console.setCookie("addCategories.js", 5);
				return targetSupport.addByCategory(categories)




					.catch(x => Q.reject({ error: 'Installation failure', details: x }))
					.finally(x => {
						targetSupport.removeListener(progresshdl);
						//gc.console.setCookie("addCategories.js", 0);
					});
			}

			// Get TICA, TargetSupport, and call addCategories to install driver
			// NOTE: I only work in the online environment TODO: Test the offine app implementation 
			if (!gc.desktop.isDesktop()) {

				console.log("Installing DFU driver...");

				return TICloudAgent.Init().then(tica => {
					return tica.getSubModule('TargetSupport');
				}).then(targetSupport => {
					// Download and install DFU driver. Filepaths are as follows:
					// Online: %LOCALAPPDATA%\Texas Instruments\TICloudAgent\loaders\ccs_base\emulation\drivers\stellaris\icdi
					// Offline: ?
					return addCategories(targetSupport, ['lm_icdi']); // to install the DFU driver
				}).then(() => { console.log("Installation complete"); });

			} else {
				return Q();
			}
		}

		// In your index.gui file, add the following lines ti include this tool
		// 1) <link rel="import" href="../ti-tool-bsl/ti-tool-bsl.html">
		// 2) <ti-tool-bsl></ti-tool-bsl>
		Utilities.bsl_update = async function (scriptName, hexName) {

			try {

				// Get a reference to the BSL Scriptor tool
				let bslTool = document.querySelector('ti-tool-bsl');
				if (!bslTool) { throw "Could not find a reference to the BSL tool!"; }

				// Default filenames
				scriptName = scriptName || "bsl_script.txt";
				hexName = hexName || "firmware.txt";

				// Build file urls
				// NOTE: The above files should be included in the 'target/' folder!
				//var prefix = location.href;
				//if (prefix.endsWith('index.html')) { prefix = prefix.substring(0, prefix.indexOf('index.html')); }
				//prefix += 'target/';
				let prefix = Polymer.ResolveUrl.resolveUrl('./target/');
				const scriptfile_url = prefix + scriptName;
				const hexfile_url = prefix + hexName;

				// Check if files exist
				let [scriptFileResponse, hexFileResponse] = await Promise.all([fetch(scriptfile_url), fetch(hexfile_url)]);
				if (!hexFileResponse.ok) { throw `Hex file not found: ${hexfile_url}`; }
				if (!scriptFileResponse.ok) { throw `Script file not found: ${scriptfile_url}`; }

				// Suppress connection error messages
				var popup = document.querySelector('ti-device-identify-popup');
				popup.suppressAllErrorPopups = true;

				// Open up loading screen
				var loadingScreen = document.querySelector('ti-sar-loading-screen');
				loadingScreen.showScreen = true;

				// Disable auto-connect
				var identifier = document.querySelector('ti-device-identifier');
				identifier.autoDetectDisabled = true;

				let runBSL = async () => {

					console.log("Running BSL Scriptor Tool...");
					let output = await Utilities.wrapQ(bslTool.customLoad(scriptfile_url, [hexfile_url]));
					let success = output.stdout.indexOf("[ERROR_MESSAGE]") === -1;
					console.log(output.stdout);

					// CONSIDER: Checking if the DFU device is still present.

					// Close loading screen, show result dialog
					loadingScreen.showScreen = false;
					popup.showFlashFirmwareResult(success);

					// Reload the application after timeout
					setTimeout(() => { location.reload(); }, 3000);
				}

				// Delay to allow time for DFU device to enumerate
				const ENUMERATION_DELAY_MS = 5000;
				setTimeout(() => {
					let bulkDeviceKeys = Utilities.getListOfDfuDeviceKeys();
					if (bulkDeviceKeys.length === 1) { runBSL(); }
					else {
						// TODO: Give the user a better error message than this...
						console.error("DFU device not found!");
						loadingScreen.showScreen = false;
						popup.showFlashFirmwareResult(false);

						// Reload the application after timeout
						setTimeout(() => { location.reload(); }, 3000);
					}
				}, ENUMERATION_DELAY_MS);

			}
			catch (err) { console.error("Utilities.bsl_update: " + err); }
		};

		// Helper function to fire events
		Utilities.fireEvent = function (eventName, module) {

			// Check if module name was provided, if so use the GC log
			if (module) {
				gc.console.log(module, "Firing '" + eventName + "' event");
			} else {
				console.log("Firing '" + eventName + "' event");
			}
			var event = new Event(eventName);
			document.dispatchEvent(event);
		};

		// Creates a one-time only event listener that is removed after firing
		Utilities.oneTimeEvent = function (target, eventType, callback) {
			if (!target) { console.error("oneTimeEvent target is missing"); }
			let intermediateCallback = () => {
				target.removeEventListener(eventType, intermediateCallback);
				callback();
			};
			target.addEventListener(eventType, intermediateCallback);
		};

		// Creates a temporary event listener. Listener is removed if callback function returns 'true'.
		Utilities.temporaryEvent = function (target, eventType, callback) {
			if (!target) { console.error("temporaryEvent target is missing"); }
			let intermediateCallback = () => {
				let removeListener = callback();
				if (removeListener) { target.removeEventListener(eventType, intermediateCallback); }
			};
			target.addEventListener(eventType, intermediateCallback);
		};

		// Read data from JSON file
		Utilities.readJSONFile = function (filePath) {
			return gc.fileCache.readJsonFile(filePath);
		};

		// Read data from CSV file
		Utilities.readCSVFile = function (filePath) {
			return new Promise(function (resolve, reject) {
				gc.fileCache.readTextFile(filePath).then(data => {
					data = data.replace(/[\n\r]/g, '\n');
					// TODO: Convert to array
					resolve(csvFile);
				});
			});
		};

		Utilities.isEmpty = function (obj) {
			for (var key in obj) {
				if (obj.hasOwnProperty(key))
					return false;
			}
			return true;
		};

		// Tests if two objects have the same key/value pairs
		// Reference: https://dmitripavlutin.com/how-to-compare-objects-in-javascript/
		Utilities.shallowEqual = function (object1, object2) {
			const keys1 = Object.keys(object1);
			const keys2 = Object.keys(object2);

			if (keys1.length !== keys2.length) { return false; }

			for (let key of keys1) {
				if (object1[key] !== object2[key]) { return false; }
			}

			return true;
		};

		Utilities.setStatusBarProgress = function (percent) {
			let statusBar = document.getElementById("ti_widget_statusbar");
			statusBar.progress = percent;
			statusBar.progressBarEnabled = true;    //NOTE: Statusbar automatically hides if percent is 0 or 100.
		};

		Utilities.findFieldObjectById = function (fieldIdString, registerIdString, regblockIdString) {
			// // REFERENCE: https://confluence.itg.ti.com/pages/viewpage.action?pageId=202419411

			// // EXAMPLE: How to binding a dropdown menu to a register field

			//     // we then need to ask it for the JSON object that corresponds to the values in the register.json file:
			//     var registerModel = document.querySelector('ti-model-register');
			//     registerModel.getRegisterViewInformation().then(function(deviceObj) {

			//         // filter the register blocks based on id
			//         const regBlockArray = deviceObj.regblocks.filter((regblock) => (regblock.id === "device_configuration"));
			//         if (regBlockArray.length > 0) {
			//             const regBlock = regBlockArray[0];
			//             // filter the registers from that register block based on id
			//             const regArray = regBlock.registers.filter((reg) => (reg.id === "gain1_register"));
			//             if (regArray.length > 0) {
			//                 const reg = regArray[0];
			//                 // filter the bitfields from that register based on id
			//                 const bitfieldArray = reg.fields.filter((bitfield) => (bitfield.id === "pgagain0"));

			//                 if (bitfieldArray.length > 0) {
			//                     const bitfield = bitfieldArray[0];

			//                 }
			//             }
			//         }
			//     });
		};


		Utilities.parseSystemJson = function (data, deviceName, controllerName, controllerConfigName) {

			// Get device object
			let device;
			deviceName = deviceName | data.deviceDefault;
			if (!deviceName || !data.deviceMap[deviceName]) {
				console.error(`No matching device for '${deviceName}' in system.json`);
				return;
			}
			device = data.deviceMap[deviceName];

			// Get controller object
			let controller;
			controllerName = controllerName | device.controllerDefault;
			if (!controllerName || !device.controllerMap[controllerName]) {
				console.error(`No matching controller for '${deviceName}.${controllerName}' in system.json`);
				return;
			}
			controller = device.controllerMap[controllerName];

			// Get configuration object
			let config;
			controllerConfigName = controllerConfigName | controller.configurationDefault;
			if (!controllerConfigName || !controller.configurationMap[controllerConfigName]) {
				console.error(`No matching configuration for '${deviceName}.${controllerName}.${controllerConfigName}' in system.json`);
				return;
			}
			config = controller.configurationMap[controllerConfigName];

			return {
				"device": device,
				"controller": controller,
				"config": config
			};

		},

			// Reference: https://italonascimento.github.io/applying-a-timeout-to-your-promises/
			Utilities.promiseTimeout = function (ms, inputPromise) {

				// Create a promise that rejects in <ms> milliseconds
				var id;
				let timeoutPromise = new Promise((resolve, reject) => {
					id = setTimeout(() => {
						reject(`Promise timed out in ${ms} ms.`);
					}, +ms);
				});

				// Returns a race between our timeout and the passed in promise
				let racePromise = Promise.race([inputPromise, timeoutPromise]).then(result => {
					clearTimeout(id);
					return result;
				});

				return racePromise;
			};

		// Loads and executes a script file
		// Utilities.executeScript = function (scriptName, model) {

		// 	// Set default model if none is provided
		// 	model = model || gc.databind.registry.getModel('my_streaming_model');

		// 	if (gc.desktop.isDesktop()) {

		// 		//let 

		// 	}


		// 	// Read in script and bootloader files
		// 	Q.all([
		// 		gc.fileCache.readTextFile('app/scripts/' + scriptName),
		// 		gc.fileCache.readTextFile('app/scripts/macros.js')
		// 	]).then(([script, logfile, bootloader]) => {

		// 		let bufferLength = null;
		// 		const scripting = model.newScriptInstance(bufferLength);

		// 		scripting.load(script, bootloader);
		// 		scripting.main();
		// 	}).fail(e => {
		// 		console.error(e);
		// 	});
		// };

		// TODO: Create a helper function for binding drop-down list widgets to register fields - See if ti-gc-bindings-behaviors has provided a means to do this...
		// Utilities.bindDropDownListToFieldValue = function (dropDownlistObj, bitfieldObj) {

		// 	// The widget's binding to the register bitfield controls what the selected value and label is
		// 	const widgetOptions = bitfieldObj.widget.options;   // TODO: In the future, check for options on the bitfield object directly (not in the widget)

		// 	// Use the options for the bitfield to create the labels and values for the droplist
		// 	const labels = widgetOptions.map((option) => (option.display));
		// 	dropDownlistObj.labels = labels.join("|");

		// 	const values = widgetOptions.map((option) => (option.value));
		// 	dropDownlistObj.values = values.join(",");

		// 	// TODO: Add to <ti-sar-bindings> if a similar function does not exist
		// 	// The goal is to make it easier to bind a drop-down menu to a enumerated register field
		// 	// The register model reads in the .json file when it starts up, so in order to get
		// 	// access to this information, we need to wait for the register model to send a registerDefsChanged
		// 	// event that says it has updated its register model:
		// 	//var registerModel = document.querySelector('ti-model-register');
		// 	//registerModel.addEventListener('registerDefsChanged', function(event) {

		// 	// EXAMPLE: How to binding a dropdown menu to a register field
		// 	// REFERENCE: https://confluence.itg.ti.com/pages/viewpage.action?pageId=202419411
		// 	// we then need to ask it for the JSON object that corresponds to the values in the register.json file:
		// 	// registerModel.getRegisterViewInformation().then(function(deviceObj) {

		// 	//     // filter the register blocks based on id
		// 	//     const regBlockArray = deviceObj.regblocks.filter((regblock) => (regblock.id === "device_configuration"));
		// 	//     if (regBlockArray.length > 0) {
		// 	//         const regBlock = regBlockArray[0];
		// 	//         // filter the registers from that register block based on id
		// 	//         const regArray = regBlock.registers.filter((reg) => (reg.id === "gain1_register"));
		// 	//         if (regArray.length > 0) {
		// 	//             const reg = regArray[0];
		// 	//             // filter the bitfields from that register based on id
		// 	//             const bitfieldArray = reg.fields.filter((bitfield) => (bitfield.id === "pgagain0"));

		// 	//             if (bitfieldArray.length > 0) {
		// 	//                 const bitfield = bitfieldArray[0];

		// 	//                 // use the options for the bitfield to create the labels and values for the droplist
		// 	//                 const widgetOptions = bitfield.widget.options;

		// 	//                 const labels = widgetOptions.map((option) => (option.display));
		// 	//                 templateObj.$.ti_widget_droplist_demo.labels = labels.join("|");

		// 	//                 const values = widgetOptions.map((option) => (option.value));
		// 	//                 templateObj.$.ti_widget_droplist_demo.values = values.join(",");
		// 	//                 // the widget's binding to the register bitfield controls what the selected value and label is
		// 	//             }
		// 	//         }
		// 	//     }
		// 	// });
		// 	//});

		// };

		// Given a regex string (not regex object) and test sting, return a boolean indicating if they match.
		// Optionally you can specify if you want the string matching to be case senstive, by default it is NOT.
		Utilities.regexTest = function (regExString, testString, caseSensitive) {

			// Construct regular expression
			const ignoreCase = !caseSensitive || true;
			let flags = ignoreCase ? "i" : "";
			const regex = new RegExp(regExString, flags);

			// Test
			return regex.test(testString);
		};

		Utilities.saveTextFileAs = function (textData, fileFilter) {

			let templateObj = document.querySelector('#template_obj');

			if (gc.desktop.isDesktop()) {

				fileFilter = fileFilter || "";
				let fileName = "newTextFile" + fileFilter;

				gc.File.browseAndSave(textData, fileName, fileFilter, function (result, errorInfo) {
					if (errorInfo) {
						templateObj.logMsgAndNotify("error", calledBy + " " + errorInfo, false, true);
					} else if ((result) && (result.localPath)) {
						templateObj.$.ti_widget_eventlog_view.log("info", "Text file saved to " + result.localPath);
					}
				});

			} else {
				let mime = 'text/plain;charset=utf-8';
				var blob = new Blob([text], {
					type: mime
				});
				saveAs(blob, fname);
				templateObj.$.ti_widget_eventlog_view.log("info", calledBy + ": " + fileName + " file downloaded");
			}
		};

		// Wraps Q promise in native JS promise (allows the use of async/await for Q promises)
		Utilities.wrapQ = function (promise) {
			return new Promise((resolve, reject) => promise.then(resolve).fail(reject));
		};

		// Wraps native JS promise in Q promise (to return Q promise to functions that use them)
		// NOTE: You can specify a sub-property name to resolve to, for example in cases where sendCommand
		// returns an object but the caller only needs a specific value from this object 
		Utilities.wrapJS = function (promise, resultSubProperty) {
			var deferred = Q.defer();
			promise.then(function (result) {
				// Get sub-property from return element
				if (resultSubProperty) {
					let array = resultSubProperty.split(".");
					array.forEach(element => { result = result[element]; });
				}
				deferred.resolve(result);
			}).catch(function (err) {
				deferred.reject(err);
			});

			return deferred.promise;
		}

		Utilities.Queue = function () {
			collection = [];
			this.print = function () {
				console.log(collection);
			};
			this.enqueue = function (element) {
				collection.push(element);
			};
			this.dequeue = function () {
				return collection.shift();
			};
			this.front = function () {
				return collection[0];
			};
			this.size = function () {
				return collection.length;
			};
			this.isEmpty = function () {
				return (collection.length === 0);
			};
		};
	};

	templateObj = document.querySelector('#template_obj');
	if (templateObj) {
		init();
	} else {
		document.addEventListener('DOMContentLoaded', init.bind(this));
	}

})();